#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from typing import Callable

from freighter import (
    AsyncMiddleware,
    AsyncNext,
    Context,
    Middleware,
    Next,
    Payload,
    UnaryClient,
)

from synnax.user.payload import UserPayload


class InsecureCredentials(Payload):
    username: str
    password: str


class TokenResponse(Payload):
    token: str
    user: UserPayload


AUTHORIZATION_HEADER = "Authorization"
TOKEN_REFRESH_HEADER = "Refresh-Token"


def token_middleware(
    token_provider: Callable[[], str], set_token: Callable[[str], None]
) -> Middleware:
    def mw(ctx: Context, _next: Next):
        ctx.set(AUTHORIZATION_HEADER, "Bearer " + token_provider())
        out_ctx, exc = _next(ctx)
        maybe_refresh_token(out_ctx, set_token)
        return out_ctx, exc

    return mw


def async_token_middleware(
    token_provider: Callable[[], str], set_token: Callable[[str], None]
) -> AsyncMiddleware:
    async def mw(ctx: Context, _next: AsyncNext):
        ctx.set(AUTHORIZATION_HEADER, "Bearer " + token_provider())
        out_ctx, exc = await _next(ctx)
        maybe_refresh_token(out_ctx, set_token)
        return out_ctx, exc

    return mw


def maybe_refresh_token(
    ctx: Context,
    set_token: Callable[[str], None],
) -> None:
    refresh = ctx.get(TOKEN_REFRESH_HEADER, None)
    if refresh is not None:
        set_token(refresh)


class AuthenticationClient:
    _ENDPOINT = "/auth/login"

    client: UnaryClient
    username: str
    password: str
    token: str
    user: UserPayload

    def __init__(
        self,
        transport: UnaryClient,
        username: str,
        password: str,
    ) -> None:
        self.client = transport
        self.username = username
        self.password = password

    def authenticate(self) -> None:
        res, exc = self.client.send(
            self._ENDPOINT,
            InsecureCredentials(username=self.username, password=self.password),
            TokenResponse,
        )
        if exc is not None:
            raise exc
        assert res is not None
        self.token = res.token
        self.user = res.user

    def get_token(self) -> str:
        return self.token

    def set_token(self, token: str) -> None:
        self.token = token

    def middleware(self) -> list[Middleware]:
        return [token_middleware(self.get_token, self.set_token)]

    def async_middleware(self) -> list[AsyncMiddleware]:
        return [async_token_middleware(self.get_token, self.set_token)]



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import json
from dataclasses import dataclass
from enum import Enum

import freighter

_FREIGHTER_EXCEPTION_PREFIX = "sy."


@dataclass
class Field:
    field: str
    message: str


class ValidationError(Exception):
    """
    Raised when a validation error occurs.
    """

    TYPE = _FREIGHTER_EXCEPTION_PREFIX + "validation"


class FieldError(ValidationError):
    field: str

    TYPE = _FREIGHTER_EXCEPTION_PREFIX + "validation.field"

    def __init__(self, field: str, message: str):
        self.field = field
        super().__init__(f"{field}: {message}")


class AuthError(Exception):
    """
    Raised when an authentication error occurs.
    """

    TYPE = _FREIGHTER_EXCEPTION_PREFIX + "auth"

    pass


class UnexpectedError(Exception):
    """
    Raised when an unexpected error occurs.
    """

    TYPE = _FREIGHTER_EXCEPTION_PREFIX + "unexpected"

    pass


class ContiguityError(Exception):
    """
    Raised when time-series data is not contiguous.
    """

    TYPE = _FREIGHTER_EXCEPTION_PREFIX + "contiguity"

    pass


class QueryError(Exception):
    """
    Raised when a query error occurs, such as an item not found.
    """

    TYPE = _FREIGHTER_EXCEPTION_PREFIX + "query"

    pass


class NotFoundError(QueryError):
    """
    Raised when a query returns no results.
    """

    TYPE = _FREIGHTER_EXCEPTION_PREFIX + "query.not_found"

    pass


class MultipleFoundError(QueryError):
    """
    Raised when a query that should return a single result returns multiple.
    """

    TYPE = _FREIGHTER_EXCEPTION_PREFIX + "query.multiple_results"

    pass


class RouteError(Exception):
    """
    Raised when an API routing error occurs, such as a 404.
    """

    TYPE = _FREIGHTER_EXCEPTION_PREFIX + "route"

    path: str

    def __init__(self, path: str, *args):
        super().__init__(*args)
        self.path = path


def _decode(encoded: freighter.ExceptionPayload) -> Exception | None:
    if not encoded.type.startswith(_FREIGHTER_EXCEPTION_PREFIX):
        return None

    if encoded.type.startswith(AuthError.TYPE):
        return AuthError(encoded.data)

    if encoded.type.startswith(UnexpectedError.TYPE):
        return UnexpectedError(encoded.data)

    if encoded.type.startswith(ValidationError.TYPE):
        if encoded.type.startswith(FieldError.TYPE):
            values = encoded.data.split(":")
            if len(values) != 2:
                return UnexpectedError(encoded.data)
            return FieldError(values[0], values[1])
        return ValidationError(encoded.data)

    if encoded.type.startswith(QueryError.TYPE):
        if encoded.type.startswith(NotFoundError.TYPE):
            return NotFoundError(encoded.data)
        if encoded.type.startswith(MultipleFoundError.TYPE):
            return MultipleFoundError(encoded.data)
        return QueryError(encoded.data)

    if encoded.type.startswith(RouteError.TYPE):
        return RouteError(encoded.data)

    return UnexpectedError(encoded.data)


def _encode(err: Exception) -> freighter.ExceptionPayload | None:
    raise NotImplemented


freighter.register_exception(_encode, _decode)
